/* Helsingin Yliopiston TTK-91 esimerkkikonetta jotenkuten emuloiva
 * ohjelma. (c) Juho Snellman (jsnell@iki.fi) 1998. */

/* This program is under the Gnu General Public License. For more detailed
 * licensing information see the file COPYING included with the 
 * distribution */

#define MAIN

#include "ttkvm.h"
#include <getopt.h>

/* sscanf:n *-optio on epjrkev, joten kytetn &s_tmp:t sen 
 * korvikkeena */
char s_tmp;

int main (int argc, char **argv)
{
    char s[255];
    int a, c, i, count = 0, ccount, tmp, parse_only = 0,
    show_parse = 0;
    FILE *f;
    int break_p[MAXBREAKPOINT], breakcount = 0;
    
    /* Optiot defaulttitiloihin */
    print_stack = 0;
    print_inst = 0;
    print_reg = 0;
    
    for (i = 0; i < MAXBREAKPOINT; i++)
      break_p[i] = -1;
    
    while (1)
    {
	c = getopt(argc, argv, "b:hinprs");
	if (c == -1)
	  break;
	
	switch (c)
	{
	 case 'b':
	    if ((sscanf(optarg, "%d%c", &tmp, &s_tmp) != 1) || tmp < 0)
	    {
		fprintf(stderr, "Invalid breakpoint (%s).\n", optarg);
		return -1;				
	    }
	    if (breakcount >= MAXBREAKPOINT)
	    {
		fprintf(stderr, "Too many breakpoints defined.\n");
		return -1;
	    }	
	    for (i = 0; i < breakcount; i++)
	    {
		if (tmp == break_p[i])
		{
		    fprintf(stderr, "Multiple breakpoints point to same address.\n");
		    return -1;
		}
	    }
	    break_p[breakcount++] = tmp;
	    break;
	 case 'h':
	    print_usage();
	    return -1;
	    break;
	 case 'i':
	    print_inst = 1;
	    break;
	 case 'n':
	    parse_only = 1;
	    break;
	 case 'p':
	    show_parse = 1;
	    break;
	 case 'r':
	    print_reg = 1;
	    break;
	 case 's':
	    print_stack = 1;
	    break;
	 default:
	    return -1;
	}
    }
    
    if (argc == optind)
    {		
	f = fdopen(0, "r");
    }
    else if (argc > optind+1)
    {	 	
	print_usage();
	return -1;
    }
    else
    {
	f = fopen(argv[optind], "r");
    }
    
    if (f == NULL)
    {
	fprintf(stderr, "Error opening %s: %s\n", argv[optind],
		strerror(errno));
	return -1;
    }
    
    init_ftab();
    
    for (i = 0; i < MEMSPACE; i++)
      core[i] = 0;
    for (i = 0; i < 8; i++)
      reg[i] = 0;
    
    symcount = 0;
    symqcount = 0;
    pc = 0;
    sr = 0;
    
    /* devicet */
    add_symbol(SYM_ALIAS, "CRT", D_CRT);
    add_symbol(SYM_ALIAS, "KBD", D_KBD);
    
    /* syscallit */
    add_symbol(SYM_ALIAS, "HALT", S_HALT);
    
    while ( fgets(s, sizeof(s), f) )
    {
	tmp = count;
	if ( parse_command(s, &tmp) >= 0)
	{
	    core[count++] = tmp;
	} 
    }
    ccount = count;
    
    for (i = 0; i < symcount; i++)
    {
	switch (symbols[i].type)
	{
	 case SYM_ADDR:
	    if ( symbols[i].value < -65536 )
	      symbols[i].value = count++;
	    break;
	 case SYM_REGION:
	    count += symbols[i].value;
	    symbols[i].value = count - symbols[i].value;			  
	    break;
	 case SYM_ALIAS:
	    break;
	 case SYM_NONE:
	 default:
	    fprintf(stderr, "Invalid symbol type (%d)\n", symbols[i].type);
	    break;
	}
    }
    
    reg[6] = count;
    reg[7] = count;
    
    for (i = 0; i < symqcount; i++)
    {
	a = get_symbol(symq[i].name);
	if ( !(a < -65536) )
	{
	    set_addr(a, &core[symq[i].value]);
	}
	else
	{
	    fprintf(stderr, "Symbol '%s' not found\n", symq[i].name);
	    return -1;
	}
    }
    
    for (i = 0; i < ccount; i++)
    {
	if ( get_m(core[i]) == 3)
	{
	    set_m(1, &core[i]);
	    set_addr(get_addr(core[i])+i, &core[i]);
	}
	
	if (show_parse)
	{
	    printf("%03d\t%s\tR%d,%s%d(R%d)\n", i, 
		   name_command(get_op(core[i]) ), get_rj(core[i]), 
		   name_mode(get_m(core[i]) ), get_addr(core[i]), 
		   get_ri(core[i]) );
	} 
    }
    
    if (parse_only)
      return -1;
    
    if (show_parse)
      printf("--\n");
    
    while ( 1 ) 
    {
	for (i = 0; i < breakcount; i++)
	{
	    if (break_p[i] == pc)
	      debug_mode();
	}		
	do_op(core[pc]);		
	if (pc > count || pc < 0)
	{
	    printf("Program counter out of bounds. Halting.\n");
	    break;
	}
    }
    
    return 0;
}

void debug_mode(void)
{
    char s[256];
    int addr = 0;
    
    printf("  ** Breakpoint at %d reached.\n", pc);
    while(1)
    {			
	printf("  DEBUG> ");
	fflush(stdout);
	fgets(s, 255, stdin);
	fflush(stdout);
	if (feof(stdin))
	{
	    fprintf(stderr, "End of file.\n");
	    exit (-1);
	}
	switch (s[0])
	{
	 case 'h':
	 case '?':
	    printf ("  You are currently in debug-mode. The following commands are available:\n");
	    printf ("    c     -- Continue execution.\n");
	    printf ("    h     -- Show this help.\n");
	    printf ("    p n   -- Print the instruction in memory-address 'n'.\n");
	    printf ("    p n   -- Print the contents of memory-address 'n'.\n");
	    printf ("    q     -- Quit ttkvm.\n");
	    printf ("    r     -- Print the contents of the registers.\n");
	    printf ("    s     -- Print the contents of the stack.\n");
	    break;
	 case 'c':
	    return;
	 case 'p':
	 case 'i':
	    if ((sscanf(strchr(s, ' '), "%d%[^\t \n]", &addr, &s_tmp) != 1))
	    {
		printf("Invalid address.\n");
		break;
	    }
	    if (addr < -65536 || addr > 65536)
	    {
		printf("Address out of bounds.\n");
		break;
	    }
	    while (addr < 0)
	      addr += MEMSPACE;			
	    if (s[0] == 'i')
	      show_inst(addr);
	    else
	      printf("%d:\t%d\n", addr, get_addr(core[addr]));
	    break;
	 case 'q':
	    exit(0);
	 case 'r':
	    show_reg();
	    break;
	 case 's':
	    show_stack();
	    break;
	 default:
	    printf("   Invalid command, 'h' for help.\n");
	}
    }
}

char* remove_whitespace(char *s)
{
    int i, j;
    
    if (strlen(s) == 0)
      return s;
    
    for (i = 0; i < strlen(s); i++)
      if (isspace (s[i]) || s[i] < 32)
	s[i] = ' ';
    
    for (i = 0; i< strlen(s)-1; i++)
    {
	if ( (isspace(s[i]) || s[i] == ',') && (i == 0 || isspace(s[i+1]) 
						|| s[i+1] == ',') )
	{
	    if (i == 0 || s[i+1] == ',')
	      for (j = i; j <= strlen(s); j++)
		s[j] = s[j+1];
	    else
	      for (j = i+1; j <= strlen(s); j++)
		s[j] = s[j+1];
	    i--;
	}
    }
    
    for (i = strlen(s)-1; isspace(s[i]); i--)
      s[i] = 0;
    
    return s;
}

int is_register (char *s)
{
    if (s[0] == 'R' && (s[1] >= '0' && s[1] <= '7') && s[2] == 0)
      return 1;
    if (strcmp(s, "SP") == 0)
      return 1;
    if (strcmp(s, "FP") == 0)
      return 1;
    return 0;
}

int num_register (char *s)
{
    if (s[0] == 'R' && (s[1] >= '0' && s[1] <= '7') && s[2] == 0)
      return s[1]-'0';
    if (strcmp(s, "SP") == 0)
      return 6;
    if (strcmp(s, "FP") == 0)
      return 7;
    return -1;
}

int is_command (char *s)
{
    if (strcasecmp(s, "NOP") == 0)
      return 0x00;
    if (strcasecmp(s, "STORE") == 0)
      return 0x01;
    if (strcasecmp(s, "LOAD") == 0)
      return 0x02;
    if (strcasecmp(s, "IN") == 0)
      return 0x03;
    if (strcasecmp(s, "OUT") == 0)
      return 0x04;
    if (strcasecmp(s, "ADD") == 0)
      return 0x11;
    if (strcasecmp(s, "SUB") == 0)
      return 0x12;
    if (strcasecmp(s, "MUL") == 0)
      return 0x13;
    if (strcasecmp(s, "DIV") == 0)
      return 0x14;
    if (strcasecmp(s, "MOD") == 0)
      return 0x15;
    if (strcasecmp(s, "AND") == 0)
      return 0x16;
    if (strcasecmp(s, "OR") == 0)
      return 0x17;
    if (strcasecmp(s, "XOR") == 0)
      return 0x18;
    if (strcasecmp(s, "SHL") == 0)
      return 0x19;
    if (strcasecmp(s, "SHR") == 0)
      return 0x1A;
    if (strcasecmp(s, "COMP") == 0)
      return 0x1F;
    if (strcasecmp(s, "JUMP") == 0)
      return 0x20;
    if (strcasecmp(s, "JNEG") == 0)
      return 0x21;
    if (strcasecmp(s, "JZER") == 0)
      return 0x22;
    if (strcasecmp(s, "JPOS") == 0)
      return 0x23;
    if (strcasecmp(s, "JNNEG") == 0)
      return 0x24;
    if (strcasecmp(s, "JNZER") == 0)
      return 0x25;
    if (strcasecmp(s, "JNPOS") == 0)
      return 0x26;
    if (strcasecmp(s, "JLES") == 0)
      return 0x27;
    if (strcasecmp(s, "JEQU") == 0)
      return 0x28;
    if (strcasecmp(s, "JGRE") == 0)
      return 0x29;
    if (strcasecmp(s, "JNLES") == 0)
      return 0x2A;
    if (strcasecmp(s, "JNEQU") == 0)
      return 0x2B;
    if (strcasecmp(s, "JNGRE") == 0)
      return 0x2C;
    if (strcasecmp(s, "CALL") == 0)
      return 0x31;
    if (strcasecmp(s, "EXIT") == 0)
      return 0x32;
    if (strcasecmp(s, "PUSH") == 0)
      return 0x33;
    if (strcasecmp(s, "PUSHR") == 0)
      return 0x33;
    if (strcasecmp(s, "POP") == 0)
      return 0x34;
    if (strcasecmp(s, "POPR") == 0)
      return 0x34;
    if (strcasecmp(s, "SVC") == 0)
      return 0x70;
    if (strcasecmp(s, "IRET") == 0)
      return 0x71;
    return -1;
}

char* name_command (int com)
{
    switch (com)
    {	  
     case 0x00:
	return "NOP";
     case 0x01:
	return "STORE";
     case 0x02:
	return "LOAD";
     case 0x03:
	return "IN";
     case 0x04:
	return "OUT";
     case 0x11:
	return "ADD";
     case 0x12:
	return "SUB";
     case 0x13:
	return "MUL";
     case 0x14:
	return "DIV";
     case 0x15:
	return "MOD";
     case 0x16:
	return "AND";
     case 0x17:
	return "OR";
     case 0x18:
	return "XOR";
     case 0x19:
	return "SHL";
     case 0x1A:
	return "SHR";
     case 0x1F:
	return "COMP";
     case 0x20:
	return "JUMP";
     case 0x21:
	return "JNEG";
     case 0x22:
	return "JZER";
     case 0x23:
	return "JPOS";
     case 0x24:
	return "JNNEG";
     case 0x25:
	return "JNZER";
     case 0x26:
	return "JNPOS";
     case 0x27:
	return "JLES";
     case 0x28:
	return "JEQU";
     case 0x29:
	return "JGRE";
     case 0x2A:
	return "JNLES";
     case 0x2B:
	return "JNEQU";
     case 0x2C:
	return "JNGRE";
     case 0x31:
	return "CALL";
     case 0x32:
	return "EXIT";
     case 0x33:
	return "PUSH";
     case 0x34:
	return "POP";
     case 0x70:
	return "SVC";
     case 0x71:
	return "IRET";
     default:
	fprintf(stderr, "Request for unknown opcode (0x%02x).\n", com);
	break;
    }
    return "";
}

char* name_mode (int mode)
{
    switch (mode)
    {
     case 0:
	return "=";
     case 1:
	return "";
     case 2: 
	return "@";
     case 3: 
	return "*";
     default:
	fprintf(stderr, "Request for unknown addressing mode (%d).\n", mode);
	break;
    }
    return "";
}

int is_pseudo_command(char *s)
{
    if (strcasecmp(s, "DS") == 0)
      return 1;
    if (strcasecmp(s, "EQU") == 0)
      return 1;
    if (strcasecmp(s, "DC") == 0)
      return 1;
    return 0;
}

int do_pseudo_command(char *lab, char *com, char *val)
{
    int rval;
    if (sscanf(val, "%d%c", &rval, &s_tmp) != 1)
    {
	fprintf(stderr, "Invalid constant: '%s'\n", val);
	return -1;
    }
    if (rval > 65535 || rval < -65536)
    {
	fprintf(stderr, "Invalid constant: '%s'\n", val);
	return -1;
    }
    if (strcasecmp(com, "DS") == 0)
    {
	add_symbol(SYM_REGION, lab, rval);
    }
    else if (strcasecmp(com, "EQU") == 0) 	
    {
	add_symbol(SYM_ALIAS, lab, rval);
    }
    else if (strcasecmp(com, "DC") == 0)
    {
	add_symbol(SYM_ADDR, lab, -65537);
    }
    return 0;
}

int is_address (char *s)
{
    int tmp;
    char stmp[32];
    if (is_register(s))
      return 0;
    if (sscanf(s, "%d%c", &tmp, &s_tmp) == 1)
      return 1;
    /* Korvaa 16 MAXSYMSIZE:ll... FIXME */
    if (sscanf(s, "%16[a-zA-Z0-9_]%c", stmp, &s_tmp) == 1)
      return 1;
    return 0;
}

int get_symbol (char *symbol)
{
    int i;
    for (i = 0; i < symcount; i++)
      if (strcmp(symbol, symbols[i].name) == 0)
	return symbols[i].value;
    return -65537;
}

int num_address (char *s)
{
    int tmp;
    if (sscanf(s, "%d%c", &tmp, &s_tmp) == 1)
      return tmp;
    return -65537;
}

int add_symbol (int type, char *symbol, int value)
{
    if (symcount == MAXSYMBOLS)
      return -1;
    symbols[symcount].value = value;
    symbols[symcount].type = type;
    strcpy(symbols[symcount++].name, symbol);
    return 0;
}

int add_label_queue (char *symbol, int loc)
{
    if (symqcount == MAXSYMBOLS)
      return -1;
    symq[symqcount].value = loc;
    strcpy(symq[symqcount++].name, symbol);
    return 0;
}

int parse_command(char *s, int *loc)
{
    char tstr[8][MAXLABELSIZE], label[MAXLABELSIZE], command[6];
    char *end;
    int addrat = -1, addr = 0, rj = -1 , ri = 0, m = 1;
    int i, j, sc = 0;
    
    label[0] = 0;
    command[0] = 0;	
    
    for (i = 0 ; i < 8; i++)
      tstr[i][0] = '\0';
    
    if ( (end = strchr(s, ';')) )
      *end = '\0';
    remove_whitespace(s);
    if (strlen(s) == 0)
      return -1;
    
    for (i = 0; i < 8 && sc < strlen(s); i++)
    {	
	for (j = 0; j < MAXLABELSIZE-1 && sc < strlen(s); j++)
	{
	    tstr[i][j] = s[sc++];
	    if (tstr[i][j] == '=' ||
		tstr[i][j] == '*' ||
		tstr[i][j] == '@' )
	    {
		j++;
		break;
	    }
	    if (s[sc] == ' ' ||
		s[sc] == ',' ||
		s[sc] == '(' ||
		s[sc] == ')' )
	    {
		sc++;
		j++;
		break;
	    }
	}
	tstr[i][j] = 0;
    }
    
    j = 0;
    
    if ( j < i && is_command(tstr[j]) == -1)
    {
	strncpy(label, tstr[j++], sizeof(label) );
	label[sizeof(label)-1] = '\0';
    }    
    
    if (is_pseudo_command(tstr[j]) )
    {
	if (j+1 < i )
	  do_pseudo_command(tstr[0], tstr[1], tstr[2]);
	else
	  fprintf(stderr, "Not enough arguments: '%s'\n", s);
	return -1;
    }
    if ( j < i && is_command(tstr[j]) == -1)
    {
	fprintf(stderr, "Invalid command: '%s'\n", tstr[j]);
	return -1;
    }	
    
    strncpy(command, tstr[j++], sizeof(command) );
    command[sizeof(command)-1] = '\0';

    if ( j < i && !is_register(tstr[j]) )
    {
	/* Only the second operand is meaningful. */
	if (strcasecmp(command, "jump") == 0 ||
	    strcasecmp(command, "jles") == 0 ||
	    strcasecmp(command, "jequ") == 0 ||
	    strcasecmp(command, "jgre") == 0 ||
	    strcasecmp(command, "jnles") == 0 ||
	    strcasecmp(command, "jnequ") == 0 ||
	    strcasecmp(command, "jngre") == 0)
	{
	}	
	/* Discard both arguments if present. */
	else if (strcasecmp(command, "nop") == 0) 
	{
	    j++;
	}	
	else
	{
	    fprintf(stderr, "Invalid register: '%s'\n", tstr[j]);
	    return -1;
	}	
    }
    else if ( j < i )
      rj = num_register(tstr[j++]);
    
    if ( j < i && tstr[j][0] == '@' )
    {
	j++; 
	m = 2;
    }
    else if ( j < i && tstr[j][0] == '=' )
    {
	j++;
	m = 0;
    }
    else if ( j < i && tstr[j][0] == '*' )
    {
	j++;
	m = 3;
    }

    if ( j < i && is_register(tstr[j]) )
    {
	if (m == 1)
	  m = 0; 
	else if (m == 2)
	  m = 1;
	ri = num_register(tstr[j++]);
    }
    else if ( j < i && is_address(tstr[j]) && strcasecmp(command, "nop") != 0)
    {
	if (sscanf(tstr[j], "%d", &addr) == 1)
	{
	    if (addr > 65535 || addr < -65536)
	    {
		fprintf(stderr, "Invalid constant: '%s'\n", tstr[j]);
		return -1;
	    }
	}
	addrat = j;
	addr = num_address(tstr[j++]);
    }
    else if ( j < i)
    {
	if (strcasecmp(command, "nop") != 0)
	{	    
	    fprintf(stderr, "Invalid address/register: '%s'\n", tstr[j]);
	    return -1;
	}
    }

    if (j < i)
    {
	if (ri != 0 || !is_register(tstr[j]) )
	{		  
	    if (!(strcasecmp(command, "nop") == 0))
	    {	    
		fprintf(stderr, "Invalid index register: '%s'\n", tstr[j]);
		return -1;
	    }	    
	    else
	      j++;
	}
	else
	  ri = num_register(tstr[j++]);
    }	

    if ( j != i)
    {
	fprintf(stderr, "Too many arguments: '%s' (%d/%d)\n", s, j, i);
	return -1;
    }
    
    if (label[0])
    {
	add_symbol(SYM_ADDR, label, *loc);
    }
    
    if ( addr < -65536 )
    {
	add_label_queue(tstr[addrat], *loc);
	addr = 0;
    }
    
    *loc = 0;
    set_rj(rj, loc);
    set_m(m, loc);
    set_ri(ri, loc);
    set_addr(addr, loc);
    set_op(is_command(command), loc);
    
    return 0;
}

/* int on sitten 32 bitti, joohan ? */

int get_op (int val)
{
    val >>= 24; 
    return val & 0xff; /* bitit 31-24 */
}

int get_rj (int val)
{
    val >>= 21;
    return val & 0x07; /* bitit 23-21 */
}

int get_m (int val)
{
    val >>= 19;
    return val & 0x03; /* bitit 20-19 */
}

int get_ri (int val)
{
    val >>= 16;
    return val & 0x07; /* bitit 18-16 */
}

int get_addr (int val)
{
    int addr;
    addr = val & 0xffff; /* bitit 15-0 */
    if ( addr & 0x00008000)
      return addr | 0xffff0000;
    return addr;
}

int set_op (int val, int *target)
{
    *target &= ~0xff000000; 	
    val <<= 24;
    return *target |= val; 
}

int set_rj (int val, int *target)
{
    *target &= ~0x00e00000; 
    val <<= 21; 
    return *target |= val; 
}

int set_m (int val, int *target)
{
    *target &= ~0x00180000;
    val <<= 19;
    return *target |= val; 
}

int set_ri (int val, int *target)
{
    *target &= ~0x00070000;
    val <<= 16;
    return *target |= val; 
}

int set_addr (int val, int *target)
{
    if ( val > 65535 || val < -65536 )
    {
	fprintf(stderr, "Constant out of range (%d)\n", val);
	exit(-1);
    }
    
    if ( val == -32768)
      val = 0x00008000;
    
    if ( val < 0)
      val &= 0x0000ffff;
    
    *target &= ~0x0000ffff;
    return *target |= val; 
}

int set_state (int val)
{
    switch (val)
    {
     case 'E':
	sr |= 0x01; return 0;
     case 'G':
	sr |= 0x02; return 0;
     case 'L':
	sr |= 0x04; return 0;	
     case 'O':
	sr |= 0x08; return 0;	
     case 'Z':
	sr |= 0x10; return 0;	
     case 'U':
	sr |= 0x20; return 0;	
     case 'M':
	sr |= 0x40; return 0;	
     case 'I':
	sr |= 0x80; return 0;	
     case 'S':
	sr |= 0x100; return 0;	
     case 'P':
	sr |= 0x200; return 0;	
     case 'D':
	sr |= 0x400; return 0;
     default:
	fprintf(stderr, "Invalid state-register bit\n");
    }
    return -1;
}

int unset_state (int val)
{
    switch (val)
    {
     case 'E':
	sr &= ~0x01; return 0;
     case 'G':
	sr &= ~0x02; return 0;
     case 'L':
	sr &= ~0x04; return 0;	
     case 'O':
	sr &= ~0x08; return 0;	
     case 'Z':
	sr &= ~0x10; return 0;	
     case 'U':
	sr &= ~0x20; return 0;	
     case 'M':
	sr &= ~0x40; return 0;	
     case 'I':
	sr &= ~0x80; return 0;	
     case 'S':
	sr &= ~0x100; return 0;	
     case 'P':
	sr &= ~0x200; return 0;	
     case 'D':
	sr &= ~0x400; return 0;
     default:
	fprintf(stderr, "Invalid state-register bit (%c)\n", (char) val);
    }
    return -1;
}

int get_state (int val)
{
    switch (val)
    {
     case 'E':
	return sr&0x01;
     case 'G':
	return sr&0x02; 
     case 'L':
	return sr&0x04; 	
     case 'O':
	return sr&0x08; 	
     case 'Z':
	return sr&0x10; 	
     case 'U':
	return sr&0x20; 	
     case 'M':
	return sr&0x40; 	
     case 'I':
	return sr&0x80; 	
     case 'S':
	return sr&0x100; 	
     case 'P':
	return sr&0x200; 	
     case 'D':
	return sr&0x400; 
     default:
	fprintf(stderr, "Invalid state-register bit\n");
    }
    return 0;
}

int* get_core(int addr)
{
    if (addr < -65536 || addr > 65535)
    {
	fprintf(stderr, "Out-of bounds memory access (%d)\n", addr);
	exit(-1);
    }
    while (addr < 0)
      addr += MEMSPACE;
    return &core[addr];
}

int calc_addr(char act, int a, int b)
{
    int sum;
    
    unset_state('O');
    
    if ( a > 65535 || a < -65536 )
    {
	fprintf(stderr, "Address out of range (%d)\n", a);
	exit(-1);
    }
    
    if ( b > 65535 || b < -65536 )
    {
	fprintf(stderr, "Address out of range (%d)\n", b);
	exit(-1);
    }
    
    switch (act)
    {
     case '+':
	sum = a+b; break;
     case '*':
	sum = a*b; break;
     case '-':
	sum = a-b; break;
     case '/':
	sum = a/b; break;
     case '%':
	sum = a%b; break;
     default:
	fprintf(stderr, "Internal error, action '%c'\n", act);
	exit(-1);
    }
    
    
    while ( sum > 32767)
    {
	set_state('O');
	sum -= 32767;
    }
    
    while ( sum < -32768)
    {
	set_state('O');
	sum += 32768;
    }
    
    return sum;
}

void show_stack(void)
{
    int i;
    for (i = -5; i < 0; i++)
      printf("S%03d:%03d ", reg[6]+i, 
	     get_addr(*get_core(reg[6]+i)));
    printf("SP:%03d\n", get_addr(*get_core(reg[6])) );	
}

void show_reg(void)
{
    int i;
    for (i = 0; i < 8; i++)
      printf("R%d:%02d ", i, reg[i]);
    printf("PC:%02d ", pc);
    printf("SR:%08x\n", sr);
}

void show_inst(int addr)
{
    int mem = core[addr];
    printf("%d:\t%s\tR%d,%s%d(R%d)\n", addr,
	   name_command(get_op(mem)), get_rj(mem), 
	   name_mode(get_m(mem)), get_addr(mem), 
	   get_ri(mem));
}

void print_usage(void)
{
    printf("ttkvm usage: ttkvm [options] fname.\n");
    printf("Compiles and runs a TTK-91 assembler file.\n");
    printf("\n");
    printf("-b n,\t\tbreak into the debugger whenever instruction at address 'n'\n");
    printf("\t\tis about to be executed. Multiple breakpoints can be set.\n");
    printf("-h,\t\tshow this message\n");
    printf("-i,\t\tshow the instructions under execution\n");
    printf("-n,\t\tparse the file only, do not compile or run\n");
    printf("-p,\t\tshow the parsed code\n");
    printf("-r,\t\tshow the register states before execution\n");
    printf("-s,\t\tshow the top of the stack before execution\n");
    printf("\n");
    printf("Please report any bugs to jsnell@iki.fi\n");
}

