/* This program is under the Gnu General Public License. For more detailed
 * licensing information see the file COPYING included with the 
 * distribution */

#include "ttkvm.h"

void (*ftab[256])(int, int);

void do_nop (int rj, int addr)
{
	return;
}

void do_load (int rj, int addr)
{
	reg[rj] = addr;
	return;
}

void do_store (int rj, int addr)
{
	set_addr(reg[rj], get_core(addr));
	return;
}

void do_in (int rj, int addr)
{
	int i = 65536;
	char s[256];
	if (addr == D_KBD)
	{
		do
		{			
			printf("  KBD> ");
			fflush(stdout);
			fgets(s, 255, stdin);
			fflush(stdout);
			if (feof(stdin))
			{
				fprintf(stderr, "End of file.\n");
				exit (-1);
			}
		}
		while ( sscanf(s, "%d", &i) != 1 || i < -65536 || i > 65535);
		
		reg[rj] = i;
	}
	else 
	  fprintf(stderr, "Unknown device (%d)\n", addr);
	return;
}

void do_out (int rj, int addr)
{
	if (addr == D_CRT)
	  printf("  CRT> %d\n", reg[rj] );
	else 
	  fprintf(stderr, "Unknown device (%d)\n", addr);
	return;
}

void do_add (int rj, int addr)
{
	reg[rj] = calc_addr('+', reg[rj], addr);
	return;
}

void do_sub (int rj, int addr)
{
	reg[rj] = calc_addr('-', reg[rj], addr);
	return;
}

void do_mul (int rj, int addr)
{
	reg[rj] = calc_addr('*', reg[rj], addr);
	return;
}

void do_div (int rj, int addr)
{
	reg[rj] = calc_addr('/', reg[rj], addr);
	return;
}

void do_mod (int rj, int addr)
{
	reg[rj] = calc_addr('%', reg[rj], addr);
	return;
}

void do_not (int rj, int addr)
{
	reg[rj] = ~addr;
	return;
}

void do_and (int rj, int addr)
{
	reg[rj] &= addr;
	return;
}

void do_or (int rj, int addr)
{
	reg[rj] |= addr;
	return;
}

void do_xor (int rj, int addr)
{
	reg[rj] ^= addr;
	return;
}

void do_shl (int rj, int addr)
{
	reg[rj] <<= addr;
	return;
}

void do_shr (int rj, int addr)
{
	reg[rj] >>= addr;
	return;
}

void do_comp (int rj, int addr)
{
	unset_state('E');
	unset_state('G');
	unset_state('L');
	if (reg[rj] == addr)
	  set_state('E');
	else if (reg[rj] > addr)
	  set_state('G');
	else 
	  set_state('L');
	return;
}

void do_jump (int rj, int addr)
{
	pc = addr;
	return;
}

void do_jneg (int rj, int addr)
{
	do_comp(rj, 0);
	if (get_state('L'))
	  pc = addr;
	return;
}

void do_jzer (int rj, int addr)
{
	do_comp(rj, 0);
	if (get_state('E'))
	  pc = addr;
	return;
}

void do_jpos (int rj, int addr)
{
	do_comp(rj, 0);
	if (get_state('G'))
	  pc = addr;
	return;
}

void do_jnneg (int rj, int addr)
{
	do_comp(rj, 0);
	if (get_state('E') || get_state('G') )
	  pc = addr;
	return;
}

void do_jnzer (int rj, int addr)
{
	do_comp(rj, 0);
	if (get_state('L') || get_state('G') )
	  pc = addr;
	return;
}

void do_jnpos (int rj, int addr)
{
	do_comp(rj, 0);
	if (get_state('E') || get_state('L') )
	  pc = addr;
	return;
}

void do_jles (int rj, int addr)
{
	if (get_state('L'))
	  pc = addr;
	return;
}

void do_jequ (int rj, int addr)
{
	if (get_state('E'))
	  pc = addr;
	return;
}

void do_jgre (int rj, int addr)
{
	if (get_state('G'))
	  pc = addr;
	return;
}

void do_jnles (int rj, int addr)
{
	if (get_state('E') || get_state('G') )
	  pc = addr;
	return;
}

void do_jnequ (int rj, int addr)
{
	if (get_state('L') || get_state('G') )
	  pc = addr;
	return;
}

void do_jngre (int rj, int addr)
{
	if (get_state('E') || get_state('L') )
	  pc = addr;
	return;
}

void do_push (int rj, int addr)
{
	set_addr(addr, get_core(++reg[rj]));
	
	return;
}

void do_pop (int rj, int addr)
{
	if (addr < 0 || addr > 7)
	{
		fprintf(stderr, "Invalid register in do_pop (R%d)\n", addr);
	}
	reg[addr] = get_addr(*get_core(reg[rj]--));
	
	return;
}

void do_call (int rj, int addr)
{
	do_push(rj, pc);
	do_push(rj, reg[7]);
	reg[7] = reg[rj];
	pc = addr;
	return;
}

void do_exit (int rj, int addr)
{
	reg[rj] = reg[7];
	do_pop(rj, 7);
	pc = get_addr(*get_core(reg[rj]--));
	reg[rj] -= addr;
	return;
}

void do_svc (int rj, int addr)
{
	if (addr == S_HALT)
	{
		printf("System call HALT executed, stopping.\n");
		exit (0);
	}
	else
	  fprintf(stderr, "Unknown system call (%d)\n", addr);
	return;
}

int deref_addr (int addr, int ri, int mode, int *fail)
{
	int r = reg[ri];
	if ( ri == 0)
	  r = 0;
	*fail = 0;
	switch (mode)
	{
	 case 0:
		return calc_addr('+', addr, r);
	 case 1:
		return get_addr(*get_core(calc_addr('+', addr, r) ));
	 case 2:
/*		printf("+ %d\n", get_addr(*get_core(addr)) );
		printf("+ %d\n", get_addr(*get_core(
						  get_addr(*get_core(addr))+r)) ); */
		return get_addr(*get_core(get_addr
					   (*get_core(calc_addr('+', addr, r)))));
	 default:
	}
	*fail = 1;
	return -1;
}

int deref_addr_jump (int addr, int ri, int mode, int *fail)
{
	int r = reg[ri];
	if ( ri == 0)
	  r = 0;
	*fail = 0;
	switch (mode)
	{
	 case 0:
		*fail = 1;
		return -1;
	 case 1:
		return calc_addr('+', addr, r);
	 case 2:
		return get_addr(*get_core(addr+r));
	 default:
	}
	*fail = 1;
	return -1;
}

int do_op (int mem)
{
	int addr, fail;
	if ( get_op(mem) == POP)
	{
		addr = get_ri(mem);
	}
	else if ( get_op(mem) == CALL || get_op(mem) == STORE ||
			 (get_op(mem) & 0x20 &&	!(get_op(mem) & 0xd0)) )
	{
		addr = deref_addr_jump(get_addr(mem), get_ri(mem), 
							   get_m(mem), &fail);
		if (fail != 0)
		{
			fprintf(stderr, 
					"Invalid second operand for PC-modifying opcode (%s/%s%d(R%d))\n",
					name_command(get_op(mem)), name_mode(get_m(mem)), 
					get_addr(mem), get_ri(mem));
			exit(-1);
		}
	}
	else
	{
		addr = deref_addr(get_addr(mem), get_ri(mem), get_m(mem)
						  , &fail);
	}
	
	if (print_stack)
	{
		show_stack();
	}
	if (print_reg)
	{
		show_reg();
	}
	if (print_inst)
	{
		show_inst(pc);
	}

	pc++;
	
	if (ftab[get_op(mem)] == NULL)
	  fprintf(stderr, "Unknown opcode (%02x)\n", get_op(mem));
	else
	  ftab[get_op(mem)](get_rj(mem), addr);
	return 0;
}

void init_ftab (void)
{
	int i;
	for (i = 0; i < 0xff; i++)
	  ftab[i] = NULL;
	ftab[0x00] = do_nop;
	ftab[0x01] = do_store;
	ftab[0x02] = do_load;
	ftab[0x03] = do_in;
	ftab[0x04] = do_out;
	ftab[0x11] = do_add;
	ftab[0x12] = do_sub;
	ftab[0x13] = do_mul;
	ftab[0x14] = do_div;
	ftab[0x15] = do_mod;
	ftab[0x16] = do_and;
	ftab[0x17] = do_or;
	ftab[0x18] = do_xor;
	ftab[0x19] = do_shl;
	ftab[0x1a] = do_shr;
	ftab[0x1f] = do_comp;
	ftab[0x20] = do_jump;
	ftab[0x21] = do_jneg;
	ftab[0x22] = do_jzer;
	ftab[0x23] = do_jpos;
	ftab[0x24] = do_jnneg;
	ftab[0x25] = do_jnzer;
	ftab[0x26] = do_jnpos;
	ftab[0x27] = do_jles;
	ftab[0x28] = do_jequ;
	ftab[0x29] = do_jgre;
	ftab[0x2a] = do_jnles;
	ftab[0x2b] = do_jnequ;
	ftab[0x2c] = do_jngre;
	ftab[0x31] = do_call;
	ftab[0x32] = do_exit;
	ftab[0x33] = do_push;
	ftab[0x34] = do_pop;
	ftab[0x70] = do_svc; 
/*	ftab[0x71] = do_iret; */
}
