/* This program is under the Gnu General Public License. For more detailed
 * licensing information see the file COPYING included with the 
 * distribution */

#define VERSION "0.31"

/* includes */
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

/* defines */

#define MEMSPACE 65536
#define MAXLABELSIZE 16
#define MAXSYMBOLS 256
#define MAXBREAKPOINT 32

  /* kskyt */

#define NOP		0x00
#define STORE	0x01
#define LOAD	0x02
#define IN		0x03
#define OUT		0x04
#define ADD		0x11
#define SUB		0x12
#define MUL		0x13
#define DIV		0x14
#define MOD		0x15
#define AND		0x16
#define OR		0x17
#define XOR		0x18
#define SHL		0x19
#define SHR		0x1A
#define COMP	0x1F
#define JUMP	0x20
#define JNEG	0x21
#define JZER	0x22
#define JPOS	0x23
#define JNNEG	0x24
#define JNZER	0x25
#define JNPOS	0x26
#define JLES	0x27
#define JEQU	0x28
#define JGRE	0x29
#define JNLES	0x2A
#define JNEQU	0x2B
#define JNGRE	0x2C
#define CALL	0x31
#define EXIT	0x32
#define PUSH	0x33
#define POP		0x34
#define SVC		0x70
#define IRET	0x71         /* Miks tm on ? */

  /* devicet */

#define D_CRT 1
#define D_KBD 2

  /* syscallit */

#define S_HALT 1

  /* symboli-tyypit */

#define SYM_NONE    0     /* obsolete */
#define SYM_ADDR    1
#define SYM_REGION  2
#define SYM_ALIAS   3

/* typedefs / structs */

typedef struct
{
	char name[MAXLABELSIZE];
	int type;
	int value;
}
symbol_s;

#ifdef MAIN
# define PREFIX
#else 
# define PREFIX extern
#endif

/* Muisti */
PREFIX int core[MEMSPACE];
/* Rekisterit */
PREFIX int reg[8];
PREFIX int pc;
PREFIX int sr;

/* Assemblerin symbolitaulut */
PREFIX symbol_s symbols[MAXSYMBOLS];
PREFIX symbol_s symq[MAXSYMBOLS];
PREFIX int symcount;
PREFIX int symqcount;

/* Optiot */
PREFIX int print_stack;
PREFIX int print_reg;
PREFIX int print_inst;

/* prototypes */

/* main.c */
void debug_mode (void);

char* name_mode (int);
int is_command (char*);
char* name_command (int);
int is_pseudo_command(char*);
int do_pseudo_command(char*, char*, char*);

int is_register (char*);
int num_register (char*);
int is_address (char*);
int num_address (char*);

int get_symbol (char*);
int add_symbol (int, char*, int);
int add_symbol_queue (char*, int);

char* remove_whitespace(char*);
int parse_command(char*, int*);

int get_op (int);
int get_rj (int);
int get_m (int);
int get_ri (int);
int get_addr (int);

int set_op (int, int*);
int set_rj (int, int*);
int set_m (int, int*);
int set_ri (int, int*);
int set_addr (int, int*);

int set_state (int);
int unset_state (int);
int get_state (int);

int* get_core(int);
int calc_addr(char, int, int);

void show_stack(void);
void show_reg(void);
void show_inst(int);

void print_usage(void);

/* op.c */

int do_op (int);
void init_ftab(void);
